from __future__ import annotations

import logging
from typing import TYPE_CHECKING

from vectice.api.json.model_version import ModelVersionStatus
from vectice.api.json.model_version_representation import ModelVersionRepresentationOutput, ModelVersionUpdateInput

_logger = logging.getLogger(__name__)

if TYPE_CHECKING:
    from vectice.api.client import Client


class ModelVersionRepresentation:
    def __init__(self, output: ModelVersionRepresentationOutput, client: "Client"):
        self.id = output.id
        self.project_id = output.project_id
        self.name = output.name
        self.status = output.status
        self.description = output.description
        self.technique = output.technique
        self.library = output.library
        self._client = client
        self.model_name = output.model_name

    def update(self, status: str | None = None) -> None:
        """Update the status of the Model Version from the API.

        Parameters:
            status: The new status of the model. Accepted values are EXPERIMENTATION, STAGING, PRODUCTION and RETIRED.

        Returns:
            None
        """
        if status is None:
            _logger.warning("No status update provided. Nothing to update.")
            return

        try:
            status_enum = ModelVersionStatus(status.strip().upper())
        except ValueError as err:
            accepted_statuses = ", ".join([f"{status_enum.value!r}" for status_enum in ModelVersionStatus])
            raise ValueError(f"'{status}' is an invalid value. Please use [{accepted_statuses}].") from err

        model_input = ModelVersionUpdateInput(status=status_enum.value)
        self._client.update_model(self.id, model_input)
        old_status = self.status.value
        self.status = status_enum
        _logger.info(f"Model version {self.id!r} transitioned from {old_status!r} to {self.status.value!r}.")



from __future__ import annotations

import json
import logging
import os
import random
import re
import string
import tempfile
from ast import literal_eval
from importlib.util import find_spec
from typing import TYPE_CHECKING, Any, TypedDict, TypeVar

from vectice.api.http_error_handlers import VecticeException
from vectice.autolog.autolog_asset_factory import AssetFactory
from vectice.autolog.model_library import ModelLibrary

if TYPE_CHECKING:
    from tempfile import TemporaryDirectory

    import matplotlib.pyplot as plt
    from catboost.core import CatBoost
    from keras import Model as KerasModel
    from lightgbm.basic import Booster
    from matplotlib.axes import SubplotBase
    from matplotlib.figure import Figure
    from pandas import DataFrame
    from plotly.graph_objs._figure import Figure as PltFigure
    from pyspark.sql import DataFrame as SparkDF
    from sklearn.base import BaseEstimator

    from vectice.models import Phase
    from vectice.models.resource.metadata.db_metadata import TableType

    ModelTypes = TypeVar("ModelTypes", BaseEstimator, Booster, CatBoost, KerasModel)
    TModel = TypedDict("TModel", {"variable": str, "model": ModelTypes, "library": ModelLibrary, "metrics": dict})
    DataframeTypes = TypeVar("DataframeTypes", SparkDF, DataFrame)
    TDataset = TypedDict("TDataset", {"variable": str, "dataframe": DataframeTypes, "type": TableType})


try:
    import ipynbname
    from IPython.core.getipython import get_ipython
    from IPython.core.interactiveshell import InteractiveShell
    from nbformat import NO_CONVERT, read
except ModuleNotFoundError:
    raise ModuleNotFoundError(
        "To use autolog, please install extra dependencies from vectice package using '%pip install vectice[autolog]'"
    ) from None

is_plotly = False
is_matplotlib = False
if find_spec("plotly") is not None:
    from plotly.graph_objs._figure import Figure as PltFigure

    is_plotly = True

if find_spec("matplotlib") is not None:
    import matplotlib.pyplot as plt
    from matplotlib.axes import SubplotBase
    from matplotlib.figure import Figure

    is_matplotlib = True

_logger = logging.getLogger(__name__)

NOTEBOOK_CELLS = {}


def start_listen():
    ipython = get_ipython()  # type: ignore

    def _notebook_cell_content_update():
        # Collab
        # cell_id = IP.get_parent()["metadata"]["colab"]["cell_id"]

        # Databricks
        # cell_id = IP.get_parent()["metadata"]["commandId"]
        cell_metadata = ipython.get_parent()["metadata"]  # type: ignore
        if "colab" in ipython.get_parent()["metadata"]:  # type: ignore
            cell_id = cell_metadata["colab"]["cell_id"]
        elif "commandId" in ipython.get_parent()["metadata"]:  # type: ignore
            cell_id = cell_metadata["commandId"]
        else:
            cell_id = cell_metadata["cellId"]
        cell_content = ipython.get_parent()["content"]["code"]  # type: ignore
        NOTEBOOK_CELLS[cell_id] = cell_content

    def _load_ipython_extension():
        """Load the extension in IPython."""
        ipython.events.register("post_execute", _notebook_cell_content_update)  # type: ignore

    if ipython and _load_ipython_extension not in ipython.events.callbacks["post_execute"]:  # type: ignore
        _load_ipython_extension()


####### Autolog logic
class Autolog:
    def __init__(
        self,
        phase: Phase | None,
        ipy: InteractiveShell,
        is_notebook: bool,
        create: bool = True,
        note: str | None = None,
    ):
        from vectice.services.phase_service import PhaseService

        if phase is None:
            raise ValueError("Login")

        if create is True:
            iteration = phase.create_iteration()
        else:
            iteration = PhaseService(
                phase._client  # pyright: ignore[reportPrivateUsage]
            ).get_active_iteration_or_create(phase)

        self._iteration = iteration
        self._iteration_service = iteration._service  # pyright: ignore[reportPrivateUsage]
        self._ip = ipy
        self._local_vars = self._ip.user_global_ns
        self._cell_content = self._get_notebook_cell_content() if is_notebook is True else self._get_cell_content()
        self._vectice_data = self._get_variable_matches(self._local_vars)
        self._failed_assets = []

        # Get back objects to log
        self._assets = self._get_assets()
        graphs = self._get_graphs()

        if note:
            self._iteration_service.log_comment(note)
            _logger.info(f"Note logged in iteration {self._iteration.index!r}.")

        # Log objects
        self._log_assets(self._assets)

        if graphs:
            for graph in graphs:
                try:
                    self._iteration_service.log_image_or_file(graph)
                    _logger.info(f"Graph {graph!r} logged in iteration {self._iteration.index!r}.")
                except Exception as e:
                    self._failed_assets.append({"reason": e, "asset": graph, "type": "Graph"})

        if len(self._failed_assets):
            _logger.warning("The following assets failed to log:")
        for failed_asset in self._failed_assets:
            _logger.warning(f"{failed_asset['type']} {failed_asset['asset']!r}, reason: {failed_asset['reason']}")

    def _get_variable_matches(self, local_vars: dict[str, Any]) -> list[dict[Any, Any]]:
        vectice_data = []

        for cell in self._cell_content:
            vectice_match = {}
            all_vars = {}

            vectice_match["cell"] = cell
            matches = re.findall(r"(.*?) =", cell)
            if matches:
                for variable in list(matches):
                    try:
                        all_vars[variable] = local_vars[variable]
                    except KeyError:
                        pass
            vectice_match["variables"] = all_vars
            vectice_data.append(vectice_match)
        return vectice_data

    def _get_notebook_path(self) -> str:
        try:
            nb_path = ipynbname.path()
            nb_full_path = os.path.join(os.getcwd(), nb_path.name)
        except FileNotFoundError:
            nb_full_path = None
        # VS Code notebook path
        nb_full_path = self._local_vars.get("__vsc_ipynb_file__") if not nb_full_path else nb_full_path

        if not nb_full_path:
            raise ValueError("The notebook file was not found.")
        return nb_full_path

    def _get_notebook_cell_content(self) -> list:
        try:
            nb_full_path = self._get_notebook_path()
            with open(nb_full_path) as fp:
                notebook = read(fp, NO_CONVERT)
            cells = notebook["cells"]  # pyright: ignore[reportGeneralTypeIssues]
            code_cells = [cell for cell in cells if cell["cell_type"] == "code"]  # ignore: type
            return [cell["source"] for cell in code_cells]
        except Exception:
            pass

        return list(NOTEBOOK_CELLS.values())

    def _get_collab_notebook_cell_content(self) -> list:
        from vectice.utils.ipython_hook import auth_drive, get_current_collab_notebook

        drive = auth_drive()
        if not drive:
            raise ValueError("GoogleDrive failed to authenticate.")
        _, file_id = get_current_collab_notebook()
        collab_drive_file = drive.CreateFile({"id": file_id})
        collab_notebook = json.loads(collab_drive_file.GetContentString())
        cells = collab_notebook["cells"]
        return [
            "".join(cell["source"]) for cell in cells if cell["cell_type"] == "code" and len(cell["source"]) >= 1
        ]  # ignore: type

    def _get_cell_content(self) -> list:
        """Used by autolog cell to get the content of the cell. This is used to parse for variables."""
        cell_content = self._ip.get_parent()["content"]["code"]  # pyright: ignore[reportGeneralTypeIssues]
        if cell_content is None:
            raise ValueError("Failed to get cell content.")
        return [cell_content]

    def _get_assets(self) -> list[TModel | TDataset]:
        assets = []
        for data in self._vectice_data:
            for key in data["variables"].keys():
                # skip cell inputs/outputs
                if key.startswith("_"):
                    continue
                asset = data["variables"][key]
                try:
                    asset_information = AssetFactory.get_asset_service(key, asset, data).get_asset()
                    if asset_information is not None:
                        assets.append(asset_information)
                except VecticeException:
                    pass
        return assets

    def _get_lightgbm_info(self, model: Booster) -> tuple[str, dict[str, Any]] | tuple[None, None]:
        try:
            params = {
                key: value
                for key, value in model._get_loaded_param().items()  # pyright: ignore[reportPrivateUsage]
                if value is not None
            }
            return "lightgbm", params
        except AttributeError:
            return None, None

    def _get_model_library(self, model: ModelTypes):
        if "xgboost" in str(model.__class__):
            return "xgboost"
        if "lightgbm" in str(model.__class__):
            return "lightgbm"
        return "sklearn"

    def _get_sklearn_or_xgboost_or_lgbm_info(
        self, model: BaseEstimator
    ) -> tuple[str, dict[str, Any]] | tuple[None, None]:
        try:
            library = self._get_model_library(model)
            params = {str(key): value for key, value in model.get_params().items() if value is not None}
            return library, params
        except AttributeError:
            return None, None

    def _format_keras_params(self, model: KerasModel) -> dict[str, Any]:
        params: dict[str, Any] = {}

        for i, layer in enumerate(model.layers):
            params[f"Layer-{i}"] = {
                "name": layer.name,
                "param": layer.count_params(),
                "output shape": layer.output_shape,
            }
        params["Total # of weights"] = model.count_params()

        return params

    def _get_keras_info(self, model: KerasModel) -> tuple[str, dict[str, Any]] | tuple[None, None]:
        try:
            return "keras", self._format_keras_params(model)
        except Exception:
            return None, None

    def _log_assets(
        self,
        assets: list[TModel | TDataset],
    ):
        for asset in assets:
            try:
                if "model" in asset:
                    self._log_model(asset)
                else:
                    self._log_dataset(asset)
            except Exception as e:
                self._failed_assets.append(
                    {"reason": e, "asset": asset["variable"], "type": "Model" if "model" in asset else "Dataset"}
                )

    def _get_catboost_info(self, model: CatBoost) -> tuple[str, dict[str, Any]] | tuple[None, None]:
        try:
            params = {str(key): value for key, value in model.get_all_params().items() if value is not None}
            return "catboost", params
        except AttributeError:
            return None, None

    def _get_model_params(self, model: TModel) -> tuple[str, dict[str, Any]] | tuple[None, None]:
        if model["library"] is ModelLibrary.SKLEARN:
            return self._get_sklearn_or_xgboost_or_lgbm_info(model["model"])  # pyright: ignore[reportGeneralTypeIssues]
        if model["library"] is ModelLibrary.LIGHTGBM:
            return self._get_lightgbm_info(model["model"])  # pyright: ignore[reportGeneralTypeIssues]
        if model["library"] is ModelLibrary.CATBOOST:
            return self._get_catboost_info(model["model"])  # pyright: ignore[reportGeneralTypeIssues]
        if model["library"] is ModelLibrary.KERAS:
            return self._get_keras_info(model["model"])  # pyright: ignore[reportGeneralTypeIssues]
        return None, None

    def _check_for_single_model(self):
        model_list = [asset for asset in self._assets if "model" in asset]
        if len(model_list) > 1:
            return False
        return True

    def _get_single_model_metrics(self) -> dict:
        from vectice.autolog.asset_services.metric_service import MetricService

        # get a single models metrics for notebook function
        metrics = {}
        if not self._check_for_single_model():
            return {}
        for data in self._vectice_data:
            cell_metrics = MetricService(data["cell"])._get_model_metrics(data)  # type: ignore[reportPrivateUsage]
            if cell_metrics:
                metrics.update(cell_metrics)
        return metrics

    def _log_model(self, model: TModel):
        from vectice import Model

        temp_dir: TemporaryDirectory | None = None
        temp_file_path: str | None = None
        if model["library"] is ModelLibrary.KERAS:
            graph = None

            try:
                from keras.utils import plot_model

                temp_dir = tempfile.TemporaryDirectory()
                file_name = f"{model['variable']!s}_plot.png"
                temp_file_path = rf"{temp_dir.name}\{file_name}"
                graph = plot_model(model["model"], to_file=temp_file_path, show_shapes=True, show_layer_names=False)

                if graph is None:
                    temp_file_path = None
                    _logger.info(
                        "Unable to generate the model plot. Please check the 'plot_model' function from the Keras library is working correctly. Make sure that the graphviz and pydot packages are installed and configured properly."
                    )

            except Exception as e:
                _logger.info(
                    f"Unable to generate the model plot. Ensure that the Keras library is correctly installed and up-to-date. Error details: {e}"
                )

        library, params = self._get_model_params(model)

        algorithm = str(model["model"].__class__).split(".")[-1]
        algorithm = algorithm.replace("'>", "")
        model_name = f"{self._iteration.phase.id}-{model['variable']}"
        # safety check, if no metrics and single model, get metrics
        model_metrics = model["metrics"] if model["metrics"] else self._get_single_model_metrics()
        self._iteration_service.log_model(
            Model(
                library=library,
                technique=algorithm,
                metrics=model_metrics,
                properties=params,
                name=model_name,
                predictor=model["model"],
                attachments=temp_file_path,
            )
        )
        _logger.info(f"Model {model_name!r} logged in iteration {self._iteration.index!r}.")

        if temp_dir is not None:
            temp_dir.cleanup()

    def _log_dataset(self, dataset: TDataset) -> None:
        from vectice import Dataset, DatasetType, FileResource, NoResource

        resource = self._get_dataset_resource(dataset)
        dataset_name = f"{self._iteration.phase.id}-{dataset['variable']}"
        no_resource_dataset = Dataset(
            type=DatasetType.UNKNOWN,
            resource=NoResource(
                dataframes=dataset["dataframe"],
                origin="DATAFRAME",
                type=dataset["type"],
            ),
            name=dataset_name,
        )
        if resource:
            # TODO Dataset type ?
            vec_dataset = Dataset(
                type=DatasetType.UNKNOWN,
                resource=FileResource(
                    paths=resource,
                    dataframes=dataset["dataframe"],
                ),
                name=dataset_name,
            )
        else:
            vec_dataset = no_resource_dataset
        try:
            self._iteration_service.log_dataset(vec_dataset)
        except FileNotFoundError:
            self._iteration_service.log_dataset(no_resource_dataset)
        _logger.info(f"Dataset {dataset_name!r} logged in iteration {self._iteration.index!r}.")

    def _get_dataset_resource(self, dataset: TDataset) -> str | None:
        import re

        if not self._cell_content:
            return None
        try:
            # Avoid getting stray dataset with autolog.notebook()
            match = []
            for cell in self._cell_content:
                match += re.findall(rf"(?<={dataset['variable']} = pd.read_csv\().*[^,)\n]", cell)
            if len(match) < 1:
                return None
            # TODO update regex
            # check if read csv has comma dominated arguments
            return literal_eval(match[0].split(",")[0]) if match[0].find(",") else literal_eval(match[0])
        except TypeError:
            return None

    def _log_graph(self, graph: PltFigure | SubplotBase | Figure):
        # TODO fix reuse, blank graph with autolog notebook
        temp_dir = tempfile.mkdtemp()
        file_name = "".join(random.choices(string.ascii_uppercase + string.digits, k=5))
        temp_file_path = rf"{temp_dir}\{file_name}.png"
        try:
            if is_plotly and isinstance(graph, PltFigure):
                graph.write_image(temp_file_path)
                self._iteration_service.log_image_or_file(temp_file_path)
            else:
                if isinstance(graph, SubplotBase):
                    graph.figure.savefig(temp_file_path)  # pyright: ignore[reportGeneralTypeIssues]
                if isinstance(graph, Figure):
                    graph.savefig(temp_file_path)
                elif graph is plt:
                    graph.savefig(temp_file_path)  # pyright: ignore[reportGeneralTypeIssues]
                self._iteration_service.log_image_or_file(temp_file_path)
            _logger.info(f"Graph {file_name!r} logged in iteration {self._iteration.index!r}.")
        except Exception as e:
            self._failed_assets.append({"reason": e, "asset": file_name, "type": "Graph"})

    def _get_graphs(self) -> list[str]:
        graphs = []
        for cell in self._cell_content:
            # seaborn and matplotlib support
            graphs += re.findall(r"(?<=plt.savefig\().*[^\)\n]", cell)
            # plotly support
            graphs += re.findall(r"(?<=.write_image\().*[^\)\n]", cell)
        return [literal_eval(graph) for graph in graphs]



from __future__ import annotations

import logging
import mimetypes
import os
import pickle  # nosec
import re
from contextlib import contextmanager
from io import BufferedReader, BytesIO, IOBase
from pathlib import Path
from typing import TYPE_CHECKING, Any, Dict, Union

from gql.transport.exceptions import TransportQueryError
from PIL import Image, ImageFile
from rich.console import Console
from rich.table import Table

from vectice.api.http_error_handlers import VecticeException
from vectice.api.json.iteration import IterationStatus
from vectice.api.json.step import StepOutput, StepType
from vectice.models.attachment_container import AttachmentContainer

if TYPE_CHECKING:
    from vectice.api.client import Client
    from vectice.api.json import AttachmentOutput
    from vectice.api.json.dataset_version import DatasetVersionOutput
    from vectice.api.json.model_version import ModelVersionOutput
    from vectice.models.dataset import Dataset
    from vectice.models.iteration import Iteration
    from vectice.models.model import Model
    from vectice.models.step import Step


@contextmanager
def hide_logs(package: str):
    old_level = logging.getLogger(package).level
    try:
        logging.getLogger(package).setLevel(logging.ERROR)
        yield
    finally:
        logging.getLogger(package).setLevel(old_level)


def check_read_only(iteration: Iteration):
    """Check if an iteration is completed or cancelled.

    Refreshing the iteration is necessary because in a Jupyter notebook
    its status could have changed on the backend.

    Parameters:
        iteration: The iteration to check.

    Raises:
        RuntimeError: When the iteration is read-only (completed or cancelled).
    """
    refresh_iteration = iteration._phase.iteration(iteration.index)  # pyright: ignore[reportPrivateUsage]
    if refresh_iteration._status in {  # pyright: ignore[reportPrivateUsage]
        IterationStatus.Completed,
        IterationStatus.Abandoned,
    }:
        raise RuntimeError(f"The Iteration is {refresh_iteration.status} and is read-only.")


def get_step_type(
    step_output: StepOutput,
    iteration: Iteration,
) -> Step:
    # TODO: cyclic imports
    from vectice.models.step import Step
    from vectice.models.step_dataset import StepDataset
    from vectice.models.step_image import StepImage
    from vectice.models.step_model import StepModel
    from vectice.models.step_number import StepNumber
    from vectice.models.step_string import StepString

    artifacts = step_output.artifacts
    step = Step(
        id=step_output.id,
        iteration=iteration,
        name=step_output.name,
        index=step_output.index,
        slug=step_output.slug,
        description=step_output.description,
        artifacts=artifacts,
        step_type=StepType.Step,
    )

    def _get_number(text: int | float | str):
        try:
            return float(text)
        except ValueError:
            return text

    if (
        artifacts is not None  # pyright: ignore[reportUnnecessaryIsInstance, reportUnnecessaryComparison]
        and len(artifacts) >= 1  # pyright: ignore[reportUnnecessaryIsInstance, reportUnnecessaryComparison]
    ):
        artifact = artifacts[len(artifacts) - 1]
        if artifact.dataset_version_id is not None:
            return StepDataset(step, artifact)
        if artifact.model_version_id is not None:
            return StepModel(step, artifact)
        if artifact.entity_file_id:
            image = _get_image_info(iteration, artifact.entity_file_id)
            return step if image is None else StepImage(step, image)
        if artifact.text:
            str_or_float = _get_number(artifact.text)
            return (
                StepNumber(step, str_or_float)
                if isinstance(str_or_float, float)
                else StepString(step, str(str_or_float))
            )
    return step


def _get_image_info(iteration: Iteration, entity_file_id: int):
    try:
        image = iteration._client.get_entity_file_by_id(entity_file_id)  # pyright: ignore[reportPrivateUsage]
        return image.file_name
    except TransportQueryError:
        return None


def check_image_path(path: str) -> bool:
    if _check_for_comment(path):
        return False
    try:
        check_path = Path(path).exists()
    except OSError:
        return False
    _, ext = os.path.splitext(path)
    pillow_extensions = {exten for exten in Image.registered_extensions()}
    if not check_path and ext in pillow_extensions:
        raise ValueError("Check the file path.")
    if ext not in pillow_extensions:
        return False
    return True


def check_file_path(path: str) -> bool:
    if _check_for_comment(path):
        return False
    try:
        check_image_path(path)
    except ValueError:
        raise ValueError("Check the file path.") from None
    _, ext = os.path.splitext(path)
    mime_type = mimetypes.guess_type(path)[0]
    if not mime_type and ext not in [".parquet", ".ipynb"]:
        return False
    return True


def _check_for_comment(path: str) -> bool:
    _, ext = os.path.splitext(path)
    if ext:
        return False
    return True


def _validate_image_or_file(path: str) -> BufferedReader:
    try:
        return open(path, "rb")
    except FileNotFoundError:
        raise ValueError(f"The provided file {path!r} cannot be opened. Check its format and permissions.") from None


def get_image_or_file_variables(value: str | IOBase | Image.Image) -> tuple[BufferedReader | IOBase | BytesIO, str]:
    if isinstance(value, IOBase):
        image = value
        filename = os.path.basename(value.name)  # type: ignore[attr-defined]
        return image, filename
    if isinstance(value, str):
        image = _validate_image_or_file(value)
        filename = os.path.basename(image.name)
        return image, filename
    if isinstance(value, Image.Image):  # pyright: ignore[reportUnnecessaryIsInstance]
        previous_load_truncated_images = ImageFile.LOAD_TRUNCATED_IMAGES
        previous_max_image_pixels = Image.MAX_IMAGE_PIXELS
        ImageFile.LOAD_TRUNCATED_IMAGES = True
        Image.MAX_IMAGE_PIXELS = None
        in_mem_file = BytesIO()
        value.save(in_mem_file, format=value.format)
        in_mem_file.seek(0)
        filename = os.path.basename(value.filename)  # type: ignore #TODO I doubt filename really exists in PIL: https://stackoverflow.com/questions/45087638/get-image-filename-from-image-pil
        ImageFile.LOAD_TRUNCATED_IMAGES = previous_load_truncated_images
        Image.MAX_IMAGE_PIXELS = previous_max_image_pixels
        return in_mem_file, filename

    raise ValueError("Unsupported image provided.")


def temp_print(string: str | None = None, table: Table | None = None) -> None:
    console = Console(width=120)
    if string:
        print(string)
        print()
    if table:
        console.print(table)
        print()


def convert_keys_to_camel_case(input_dict: Dict[str, Any]) -> Dict[str, Union[Any, Dict[str, Any]]]:
    camel_case_dict: Dict[str, Union[Any, Dict[str, Any]]] = {}

    for key, value in input_dict.items():
        camel_case_key = re.sub(r"_([a-z])", lambda match: match.group(1).upper(), key)
        if isinstance(value, dict):
            value = convert_keys_to_camel_case(value)
        camel_case_dict[camel_case_key] = value

    return camel_case_dict


def format_attachments(attachments: str | list[str]) -> list[str]:
    list_attachments = (
        [attachment for attachment in set(attachments)] if isinstance(attachments, list) else [attachments]
    )
    for attachment in list_attachments:
        if not isinstance(attachment, str):  # pyright: ignore[reportUnnecessaryIsInstance]
            raise ValueError(
                f"Argument 'attachments' with type '{type(attachment)}' is invalid, only str are supported."
            )
    return list_attachments


def check_string_sanity(value: str):
    if value == "":
        raise VecticeException("Cannot assign an empty comment. Please provide a valid comment")


def ensure_correct_project_id_from_representation_objs(project_id: str, value: Any):
    from vectice.models.representation.dataset_representation import DatasetRepresentation
    from vectice.models.representation.dataset_version_representation import DatasetVersionRepresentation
    from vectice.models.representation.model_representation import ModelRepresentation
    from vectice.models.representation.model_version_representation import ModelVersionRepresentation

    if (
        isinstance(value, DatasetRepresentation)
        or isinstance(value, DatasetVersionRepresentation)
        or isinstance(value, ModelRepresentation)
        or isinstance(value, ModelVersionRepresentation)
    ) and value.project_id != project_id:
        raise VecticeException("Assigning an asset coming from another project is forbidden")


def set_model_attachments(
    client: Client, model: Model, model_version: ModelVersionOutput
) -> tuple[list[AttachmentOutput] | None, bool]:
    logging.getLogger("vectice.models.attachment_container").propagate = True
    attachments = None
    if model.attachments:
        container = AttachmentContainer(model_version, client)
        attachments = container.add_attachments(model.attachments)

    success_pickle: bool = False

    if model.predictor:
        try:
            model_content = _serialize_model(model.predictor)
            model_type_name = type(model.predictor).__name__
            container = AttachmentContainer(model_version, client)
            container.add_serialized_model(model_type_name, model_content)
            success_pickle = True
        except Exception:
            success_pickle = False

    return attachments, success_pickle


def set_dataset_attachments(client: Client, dataset: Dataset, dataset_version: DatasetVersionOutput):
    logging.getLogger("vectice.models.attachment_container").propagate = True
    attachments = None
    if dataset.attachments:
        container = AttachmentContainer(dataset_version, client)
        attachments = container.add_attachments(dataset.attachments)
    return attachments


def _serialize_model(model: Any) -> bytes:
    return pickle.dumps(model)



# This captures the code called, we then use regex to identify the variables in locals. The variables found can then be logged to vectice.
"""Auto-Logging assets with Vectice.

NOTE: **This feature is currently in Beta and is subject to change. To activate this feature please contact your sales representative.**

------------
The enhanced auto-logging feature in Vectice allows for seamless documentation and management of your data science projects. Please note the following details about this beta feature.

NOTE: **This feature is designed to work specifically with notebooks.**

1.** Installation:**
    To enable the auto-logging feature, install the additional dependencies using the following command:

    ```bash
    pip install vectice[autolog]
    ```

2.** Supported libraries and environment:**
    Vectice automatically identifies and log assets encapsulated within a specified list of supported libraries and environement mentioned below

NOTE: **Supported libraries and environment**
    - Dataframe: Pandas, Spark.
    - Model: Scikit, Xgboost, Lightgbm, Catboost, Keras.
    - Graphs: Matplotlib, Seaborn, Plotly.
    - Environments: Vscode, Colab, Jupyter, Vertex notebook.

3.** General behavior:**
    Vectice autolog provides three methods: autolog.config, autolog.notebook, and autolog.cell. These methods are designed to log every asset to Vectice existing as a variable in the notebook's memory. It is important to review the specific behaviors outlined in the documentation for each of these three methods.

NOTE: **IMPORTANT INFORMATION**
    - For GRAPHS, ensure they are saved as files to be automatically logged in the iteration i.e:
        - plt.savefig("my figure.png") (for seaborn or matploltlib)
        - fig.write_image("my figure.png") (for plotly)
    - For METRICS, Vectice currently recognizes sklearn metrics for automatic association with models.
        - In cases there's ambiguity due to multiple models with different metrics, Vectice won't automatically link them. To establish the link, make sure each model and its respective metrics are placed within the same notebook cell.

"""
from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Any

from typing_extensions import TypedDict

if TYPE_CHECKING:
    from IPython.core.interactiveshell import InteractiveShell

    from vectice.models import Phase


class Login(TypedDict):
    phase: Phase | None


LOGIN: Login = {"phase": None}


_logger = logging.getLogger(__name__)


# Setup vectice connection on phass
def config(api_token: str, phase: str, host: str | None = None) -> None:
    """Configures the autolog functionality within Vectice.

    The `config` method allows you to establish a connection to Vectice and specify the phase in which you want to autolog your work.

    ```python
    # Configure autolog
    from vectice import autolog
    autolog.config(
        api_token = 'your-api-key', # Paste your api key
        phase = 'PHA-XXX'           # Paste your Phase Id
    )
    ```

    Parameters:
            api_token: The api token provided inside your Vectice app (API key).
            phase: The ID of the phase in which you wish to autolog your work as an iteration.
            host: The backend host to which the client will connect. If not found, the default endpoint https://app.vectice.com is used.

    """
    from vectice import Connection
    from vectice.connection import DEFAULT_HOST

    host = host or DEFAULT_HOST
    vec = Connection(api_token=api_token, host=host)
    client = vec._client  # pyright: ignore[reportPrivateUsage]
    asset = client.get_user_and_default_workspace()
    LOGIN["phase"] = vec.phase(phase)
    user_name = asset["user"]["name"]
    _logger.info(f"Welcome, {user_name}. You`re now successfully connected to Vectice.")
    client.assert_feature_flag_or_raise("autolog")

    from vectice.autolog.autolog_class import start_listen

    start_listen()


# Log the whole notebook inside Vectice iteration
def notebook(note: str | None = None) -> None:
    """Automatically log all supported models, dataframes, and graphs from your notebook within the Vectice App as assets.

    NOTE: **IMPORTANT INFORMATION**
        Autolog must be configured at the beginning of your notebook to capture all relevant information. Cells executed prior to configuring autolog may not have their assets recorded by the autolog.notebook() method and may need to be run again.

    ```python
                                   ...
    #Add this command at the end of notebook to log all the assets in memory
    autolog.notebook()
    ```

    NOTE: **Ensure that the required assets are in memory before calling this method.**

    Parameters:
        note: the note or comment to log to your iteration associated with the autolog.notebook
    """
    from vectice.autolog.autolog_class import Autolog

    # TODO add notebook parsing of content
    Autolog(LOGIN["phase"], _check_if_notebook(), True, True, note)


def cell(create_new_iteration: bool = False, note: str | None = None):
    """Automatically logs all supported models, dataframes, and graphs from a specific notebook cell within the Vectice platform.

    This method facilitates the selective logging of assets within a particular notebook cell, allowing users to precisely choose the assets to log to Vectice with an optional control to log assets inside a new iteration.

    NOTE: **Ensure that you have configured the autolog with your Vectice API token and the relevant phase ID before using this method.**

    ```python
                                   ...
    #Add this command at the end of the desired cell to log all the cells assets
    autolog.cell()
    ```

    NOTE: **Place the command at the end of the desired cell to log all assets within that cell.**

    Parameters:
        create_new_iteration: If set to False, logging of assets will happen in the last updated iteration. Otherwise, it will create a new iteration for logging the cell's assets.
        note: the note or comment to log to your iteration associated with the autolog.cell

    """
    from vectice.autolog.autolog_class import Autolog

    Autolog(LOGIN["phase"], _check_if_notebook(), False, create_new_iteration, note)


def cell_vars(offset: int = 1) -> dict[str, Any] | None:
    # Not used currently but this gets the content of the entire notebook and parses for variables.
    import io
    from contextlib import redirect_stdout

    from IPython.core.getipython import get_ipython

    ipy = get_ipython()
    out = io.StringIO()
    if not ipy:
        return None
    with redirect_stdout(out):
        ipy.magic("history {0}".format(ipy.execution_count - offset))

    # process each line...
    x = out.getvalue().replace(" ", "").split("\n")
    x = [a.split("=")[0] for a in x if "=" in a]  # all of the variables in the cell
    g = globals()
    result = {k: g[k] for k in x if k in g}
    return result


def _check_if_notebook() -> InteractiveShell:
    from IPython.core.getipython import get_ipython

    ipython = get_ipython()

    if ipython is None:
        raise ValueError("Not a notebook.")
    return ipython



from __future__ import annotations

from io import BufferedReader, IOBase
from typing import TYPE_CHECKING

from PIL.Image import Image

from vectice.api.http_error_handlers import VecticeException
from vectice.api.json.iteration import IterationStepArtifactInput, IterationStepArtifactType
from vectice.utils.common_utils import (
    check_string_sanity,
    get_image_or_file_variables,
    set_dataset_attachments,
    set_model_attachments,
)

if TYPE_CHECKING:
    from vectice.api.client import Client
    from vectice.models.dataset import Dataset
    from vectice.models.iteration import Iteration
    from vectice.models.model import Model


class IterationService:
    def __init__(self, iteration: Iteration, client: Client):
        self._client = client
        self._iteration = iteration

    def log_image_or_file(self, asset: str | IOBase | Image, section: str | None = None):
        _, filename = self._create_image_or_file_artifact(asset, section)
        return filename

    def log_comment(self, asset: int | float | str, section: str | None = None):
        if isinstance(asset, str):
            check_string_sanity(asset)
        artifact_inp: list[IterationStepArtifactInput] = [IterationStepArtifactInput(type="Comment", text=str(asset))]
        self._client.add_iteration_artifacts(iteration_id=self._iteration.id, artifacts=artifact_inp, section=section)

    def log_model(self, asset: Model, section: str | None = None):
        code_version_id = self._get_code_version_id()
        model_data = self._client.register_model(
            model=asset,
            phase=self._iteration.phase,
            iteration=self._iteration,
            code_version_id=code_version_id,
            section=section,
        )
        attachments_output, success_pickle = set_model_attachments(self._client, asset, model_data.model_version)
        return model_data, attachments_output, success_pickle

    def log_dataset(self, asset: Dataset, section: str | None = None):
        code_version_id = self._get_code_version_id()
        dataset_output = self._client.register_dataset_from_source(
            dataset=asset,
            iteration_id=self._iteration.id,
            code_version_id=code_version_id,
            section=section,
        )
        attachments_output = set_dataset_attachments(self._client, asset, dataset_output.dataset_version)
        return dataset_output, attachments_output

    def _create_image_or_file_artifact(
        self, value: str | IOBase | Image, section: str | None = None
    ) -> tuple[IterationStepArtifactInput, str]:
        is_image_or_file, filename = get_image_or_file_variables(value)
        try:
            artifact, *_ = self._client.create_iteration_attachments(
                files=[("file", (filename, is_image_or_file))], iteration_id=self._iteration.id, step_name=section
            )
            if isinstance(is_image_or_file, BufferedReader):
                is_image_or_file.close()
            return (
                IterationStepArtifactInput(id=artifact.fileId, type=IterationStepArtifactType.EntityFile.name),
                filename,
            )
        except VecticeException as error:
            raise error
        except Exception as error:
            raise ValueError("Check the provided image.") from error

    def _get_code_version_id(self):
        # TODO: cyclic imports
        from vectice import code_capture
        from vectice.models.git_version import check_code_source, inform_if_git_repo

        if code_capture:
            return check_code_source(self._client, self._iteration.phase.project.id)
        else:
            inform_if_git_repo()
            return None



from typing import TYPE_CHECKING

from vectice.api.json.model_representation import ModelRepresentationOutput
from vectice.models.representation.model_version_representation import ModelVersionRepresentation

if TYPE_CHECKING:
    from vectice.api.client import Client


class ModelRepresentation:
    def __init__(self, output: ModelRepresentationOutput, client: "Client"):
        self.id = output.id
        self.project_id = output.project_id
        self.name = output.name
        self.type = output.type
        self.description = output.description
        self.version = ModelVersionRepresentation(output.version, client=client)



from __future__ import annotations

import logging
from contextlib import suppress
from io import IOBase
from textwrap import dedent
from typing import TYPE_CHECKING, Any, ClassVar

from PIL.Image import Image
from rich.table import Table

from vectice.api.http_error_handlers import ArtifactVersionIdExistingError, NoStepsInPhaseError, VecticeException
from vectice.api.json.iteration import (
    IterationInput,
    IterationOutput,
    IterationStatus,
    IterationStepArtifactInput,
)
from vectice.models.step import Step
from vectice.services.iteration_service import IterationService
from vectice.utils.common_utils import (
    check_read_only,
    ensure_correct_project_id_from_representation_objs,
    get_step_type,
    temp_print,
)
from vectice.utils.deprecation import deprecate
from vectice.utils.instance_helper import is_image_or_file
from vectice.utils.last_assets import (
    assign_asset_version_logging,
    comment_or_image_logging,
    register_dataset_logging,
    register_model_logging,
)
from vectice.utils.logging_utils import get_iteration_status

if TYPE_CHECKING:
    from vectice import Connection
    from vectice.api import Client
    from vectice.models import Phase, Project, Workspace
    from vectice.models.dataset import Dataset
    from vectice.models.model import Model
    from vectice.models.representation.dataset_representation import DatasetRepresentation
    from vectice.models.representation.dataset_version_representation import DatasetVersionRepresentation
    from vectice.models.representation.model_representation import ModelRepresentation
    from vectice.models.representation.model_version_representation import ModelVersionRepresentation

    TAssetType = ModelRepresentation | ModelVersionRepresentation | DatasetRepresentation | DatasetVersionRepresentation

_logger = logging.getLogger(__name__)


class Iteration:
    """Represent a Vectice iteration.

    Iterations reflect the development and test cycles completed
    by data scientists until a fully functional algorithm is ready for
    deployment. Each iteration can be organised into sections for clarity.

    Typical usage example:

    ```python
    # Create a new iteration within a phase
    iteration = phase.create_iteration()

    # Log a dataset to the iteration
    iteration.log(my_dataset)

    # Log a model to the iteration with optional section assignment
    iteration.log(my_model, section="Modeling")
    ```

    NOTE: **You can create sections dynamically using the API.**

    To create a new iteration:

    ```python
    iteration = phase.create_iteration()
    ```
    """

    __slots__: ClassVar[list[str]] = [
        "_id",
        "_index",
        "_phase",
        "_status",
        "_client",
        "_steps",
        "_service",
        "__dict__",
    ]

    def __init__(
        self,
        output: IterationOutput,
        phase: Phase,
        client: Client,
    ):
        self.__dict__ = {}
        self._id = output.id
        self._index = output.index
        self._phase = phase
        self._status = output.status
        self._client = client
        self._steps = self._populate_steps()
        self._service = IterationService(self, self._client)

    def __repr__(self):
        return f"Iteration (index={self._index}, status={self._status}"

    def __eq__(self, other: object):
        if not isinstance(other, Iteration):
            return NotImplemented
        return self.id == other.id

    def _populate_steps(self) -> dict[str, Step]:
        with suppress(NoStepsInPhaseError):
            step_output = self._client.list_steps(self.id, True).list
            steps = [get_step_type(step_output=item, iteration=self) for item in step_output]
            check = {step.slug: step for step in steps}
            return check
        return {}

    def __getattr__(self, item: str) -> Step:
        if item in self.__dict__:
            return self.__dict__[item]
        if item in super().__getattribute__("_steps"):
            step_item = self._steps[item]
            step = self._client.get_step_by_name(step_item.name, self.id)
            return get_step_type(step_output=step, iteration=self)
        raise AttributeError(f"The attribute '{item}' does not exist.")

    def __setattr__(self, attr_name: str, attr_value: Any):
        if hasattr(self, "_steps") and attr_name in super().__getattribute__("_steps"):
            self._steps[attr_name] = self._steps[attr_name].step_factory_and_update(value=attr_value)
        elif hasattr(self, "__dict__") and attr_name not in self.__slots__:
            raise AttributeError(f"The attribute '{attr_name}' does not exist.")
        else:
            super().__setattr__(attr_name, attr_value)

    @property
    def id(self) -> str:
        """The iteration's identifier.

        Returns:
            The iteration's identifier.
        """
        return self._id

    @id.setter
    def id(self, iteration_id: str):
        """Set the iteration's identifier.

        Parameters:
            iteration_id: The identifier.
        """
        check_read_only(self)
        self._id = iteration_id

    @property
    def index(self) -> int:
        """The iteration's index.

        Returns:
            The iteration's index.
        """
        return self._index

    @property
    def status(self) -> str:
        """The iteration's status.

        Returns:
            The iteration's status.
        """
        return get_iteration_status(self._status)

    @property
    def completed(self) -> bool:
        """Whether this iteration is completed.

        Returns:
            Whether the iteration is completed.
        """
        return self._status is IterationStatus.Completed

    @property
    def properties(self) -> dict[str, int | str]:
        """The iteration's identifier and index.

        Returns:
            A dictionary containing the `id` and `index` items.
        """
        return {"id": self.id, "index": self.index}

    def list_phase_requirements(self) -> None:
        """Prints a list of phase requirements in a tabular format, limited to the first 10 requirements. A link is provided to view the remaining requirements.

        Returns:
            None
        """
        paged_requirements = self._client.list_step_definitions(self.phase.id)

        rich_table = Table(expand=True, show_edge=False)

        rich_table.add_column("Title", justify="left", no_wrap=True, min_width=5, max_width=40)
        rich_table.add_column("Description", justify="left", no_wrap=False, min_width=5, max_width=35)

        for requirement in paged_requirements.list:
            description = requirement.description.strip() if requirement.description else None
            rich_table.add_row(requirement.name, description)

        description = f"""There are {paged_requirements.total} requirements in the phase {self.phase.name!r} and a maximum of 10 requirements are displayed in the table below:"""
        link = dedent(
            f"""
                For quick access to the phase requirements in the Vectice web app, visit:
                {self._client.auth.api_base_url}/phase/{self.phase.id}/requirements
            """
        ).lstrip()

        temp_print(description)
        temp_print(table=rich_table)
        temp_print(link)

    def list_sections(self) -> None:
        """Prints a list of sections belonging to the iteration in a tabular format, limited to the first 10 sections. A link is provided to view the remaining sections.

        Returns:
            None
        """
        iteration_output = self._client.list_sections(self.id)

        rich_table = Table(expand=True, show_edge=False)

        rich_table.add_column("Name", justify="left", no_wrap=True, min_width=5, max_width=40)
        rich_table.add_column("Assets (count)", justify="left", no_wrap=True, min_width=5, max_width=15)

        if iteration_output.artifacts_count > 0:
            rich_table.add_row("None ('Assets without section')", str(iteration_output.artifacts_count))

        for section in iteration_output.sections.list:
            rich_table.add_row(section.name, str(section.artifacts_count))

        description = f"""There are {iteration_output.sections.total} sections in the iteration {self.index!r} and a maximum of 10 sections are displayed in the table below:"""
        link = dedent(
            f"""
                >> To log an asset to a specific section, use iteration.log()
                For quick access to the iteration in the Vectice web app, visit:
                {self._client.auth.api_base_url}/browse/iteration/{self.id}
            """
        ).lstrip()

        temp_print(description)
        temp_print(table=rich_table)
        temp_print(link)

    @deprecate(
        warn_at="23.4",
        fail_at="24.1",
        remove_at="24.2",
        reason="Steps are deprecated. Please use iteration.list_sections() instead.",
    )
    def list_steps(self) -> None:
        """WARNING: **DEPRECATED** To list the sections of your iteration, use list_sections() instead.

        Prints a list of steps belonging to the iteration in a tabular format, limited to the first 10 steps. A link is provided to view the remaining steps.

        Returns:
            None
        """
        steps_output = self._client.list_steps(self.id, populate=False)

        rich_table = Table(expand=True, show_edge=False)

        rich_table.add_column("Shortcut", justify="left", no_wrap=True, min_width=5, max_width=20)
        rich_table.add_column("Registered items (count)", justify="left", no_wrap=True, min_width=5, max_width=15)
        rich_table.add_column("Description", justify="left", no_wrap=True, min_width=5, max_width=35)

        if steps_output.total == 0:
            _logger.info(
                "The list of steps is empty. Use iteration.log() to log an asset to vectice.\n"
                "The += syntax is deprecated, if you still want to use it, close that iteration and create a new one with steps."
            )
            return

        for step in steps_output.list:
            description = step.description.strip() if step.description else None
            rich_table.add_row(step.slug, str(step.artifacts_count), description)
        description = f"""There are {steps_output.total} steps in Iteration '{self.index!r}' and a maximum of 10 steps are displayed in the table below:"""
        tips = dedent(
            """
        To access a specific step, use the step shortcut \033[1iiteration\033[0m.step_my_step_name
        The step reference is referred to as shortcut"""
        ).lstrip()
        link = dedent(
            f"""
        For quick access to the list of steps in the Vectice web app, visit:
        {self._client.auth.api_base_url}/browse/iteration/{self.id}"""
        ).lstrip()

        temp_print(description)
        temp_print(table=rich_table)
        temp_print(tips)
        temp_print(link)

    def cancel(self) -> None:
        """Cancel the iteration."""
        iteration_input = IterationInput(status=IterationStatus.Abandoned.name)
        self._client.update_iteration(self.id, iteration_input)
        self._status = IterationStatus.Abandoned
        _logger.info(f"Iteration with index {self.index} cancelled.")

    def complete(self) -> None:
        """Mark the iteration as completed."""
        iteration_input = IterationInput(status=IterationStatus.Completed.name)
        self._client.update_iteration(self.id, iteration_input)
        self._status = IterationStatus.Completed
        logging_output = dedent(
            f"""
                        Iteration with index {self.index} completed.

                        For quick access to the Iteration in the Vectice web app, visit:
                        {self._client.auth.api_base_url}/browse/iteration/{self.id}"""
        ).lstrip()
        _logger.info(logging_output)

    def delete(self) -> None:
        """Permanently deletes the iteration."""
        self._client.delete_iteration(self.id)
        _logger.info(f"Iteration with index {self.index} was deleted.")

    @property
    def connection(self) -> Connection:
        """The connection to which this iteration belongs.

        Returns:
            The connection to which this iteration belongs.
        """
        return self._phase.connection

    @property
    def workspace(self) -> Workspace:
        """The workspace to which this iteration belongs.

        Returns:
            The workspace to which this iteration belongs.
        """
        return self._phase.workspace

    @property
    def project(self) -> Project:
        """The project to which this iteration belongs.

        Returns:
            The project to which this iteration belongs.
        """
        return self._phase.project

    @property
    def phase(self) -> Phase:
        """The phase to which this iteration belongs.

        Returns:
            The phase to which this iteration belongs.
        """
        return self._phase

    def delete_assets_in_section(self, section: str) -> None:
        """Delete assets within a specified section of the iteration. This action is irreversible, and the deleted assets cannot be recovered.

        ```python
        iteration.delete_assets_in_section(section="Collect Initial Data")
        ```

        Parameters:
            section: The iteration's section where the assets are displayed in the Vectice App.
        """
        self._client.remove_iteration_assets(iteration_id=self._id, section=section)
        _logger.info(f"All assets in section {section!r} are deleted.")

    def delete_assets_without_a_section(self) -> None:
        """Delete assets without a section of the iteration. This action is irreversible, and the deleted assets cannot be recovered.

        ```python
        iteration.delete_assets_without_a_section()
        ```
        """
        self._client.remove_iteration_assets(iteration_id=self._id)
        _logger.info("All assets without a section are deleted.")

    def delete_assets_from_iteration(self) -> None:
        """Delete all assets from the iteration. This action is irreversible, and the deleted assets cannot be recovered.

        ```python
        iteration.delete_assets_from_iteration()
        ```
        """
        self._client.remove_iteration_assets(iteration_id=self._id, all=True)
        _logger.info(f"All assets from iteration {self.index!r} are deleted.")

    def log(self, asset: Any, section: str | None = None) -> None:
        """Log an asset to an iteration. Assets can be organized into sections for improved clarity within the iteration.

        ```python
        from vectice import Model
        my_model = Model(name="my_model")
        iteration.log(my_model)
        iteration.log("my note")
        ```

        Parameters:
            asset: The asset to log to an iteration. Assets can include notes, images, Vectice datasets, or Vectice models.
            section (Optional): The iteration's section where the asset will be displayed in the Vectice App. Assets without sections are logged in the iteration itself.

        """
        # TODO: cyclic imports
        from vectice.models.dataset import Dataset
        from vectice.models.model import Model
        from vectice.models.representation.dataset_representation import (
            DatasetRepresentation,
            DatasetVersionRepresentation,
        )
        from vectice.models.representation.model_representation import ModelRepresentation, ModelVersionRepresentation

        ensure_correct_project_id_from_representation_objs(self.project.id, asset)

        if section:
            section = section.strip()

        if section is not None and len(section) > 60:
            raise VecticeException("section: name must be shorter than or equal to 60 characters")

        if is_image_or_file(asset):
            self._log_image_or_file(asset, section)

        elif isinstance(asset, (int, float, str)):
            self._log_comment(asset, section)

        elif isinstance(
            asset,
            (ModelRepresentation, ModelVersionRepresentation, DatasetRepresentation, DatasetVersionRepresentation),
        ):
            self._assign_version_representation(asset, section)

        elif isinstance(asset, Model):
            self._log_model(asset, section)

        elif isinstance(asset, Dataset):
            self._log_dataset(asset, section)
        else:
            raise TypeError(f"Expected Image, Comment, Dataset or a Model, got {type(asset)}")

    def _log_image_or_file(self, asset: str | IOBase | Image, section: str | None = None):
        filename = self._service.log_image_or_file(asset, section)
        comment_or_image_logging(self, _logger, section, filename)

    def _log_comment(self, asset: int | float | str, section: str | None = None):
        self._service.log_comment(asset, section)
        comment_or_image_logging(self, _logger, section)

    def _assign_version_representation(
        self,
        asset: TAssetType,
        section: str | None = None,
    ):
        from vectice.models.representation.dataset_representation import DatasetRepresentation
        from vectice.models.representation.model_representation import ModelRepresentation

        asset_type = self._get_asset_type_from_asset_representation(asset)

        artifact_inp = [
            IterationStepArtifactInput(
                type=asset_type,
                id=asset.version.id if isinstance(asset, (ModelRepresentation, DatasetRepresentation)) else asset.id,
            )
        ]
        already_assigned: bool = False
        try:
            self._client.add_iteration_artifacts(iteration_id=self._id, artifacts=artifact_inp, section=section)
        except ArtifactVersionIdExistingError:
            already_assigned = True
        assign_asset_version_logging(
            iteration=self,
            value=asset,
            _logger=_logger,
            already_assigned=already_assigned,
            section=section,
        )

    def _log_model(self, asset: Model, section: str | None = None):
        model_data, attachments_output, success_pickle = self._service.log_model(asset, section)
        register_model_logging(self, model_data, asset, attachments_output, success_pickle, _logger, section)

    def _log_dataset(self, asset: Dataset, section: str | None = None):
        dataset_output, attachments_output = self._service.log_dataset(asset, section)
        register_dataset_logging(self, dataset_output, asset, attachments_output, _logger, section)

    def _get_asset_type_from_asset_representation(
        self,
        asset: TAssetType,
    ):
        from vectice.models.representation.dataset_representation import (
            DatasetRepresentation,
            DatasetVersionRepresentation,
        )
        from vectice.models.representation.model_representation import ModelRepresentation, ModelVersionRepresentation

        if isinstance(asset, (ModelRepresentation, ModelVersionRepresentation)):
            return "ModelVersion"
        if isinstance(
            asset, (DatasetRepresentation, DatasetVersionRepresentation)
        ):  # pyright: ignore[reportUnnecessaryIsInstance]
            return "DataSetVersion"

        raise ValueError(f"Asset type '{type(asset)!s}' is unsupported")



from __future__ import annotations

import json
import os
import re
import subprocess
import urllib.error
import urllib.request
from pathlib import Path, PurePath
from typing import Generator
from urllib.parse import unquote

import requests
from pydrive2.drive import GoogleDrive

from vectice.api.http_error import HttpError


def get_absolute_path() -> str:
    """Get the path for the directory walk to check if there's a .git and initialize the Repo object.

    Relates to 'drive.mount('/content/drive')' as '/content/' is part of the path and 'drive/' is
    required or the path will not be found.

    """
    filename, file_id = get_current_collab_notebook()
    path = str(get_collab_path(file_id=file_id, drive=None))
    try:
        path_formatting = path.split("/")
        path_formatting.pop()
        completed_path = "/".join(path_formatting)
        return f"drive/{completed_path}/{filename}"
    except ValueError:
        return str(path)


def get_current_collab_notebook():
    """Get the filename, file_id of this notebook."""
    jupyter_address = _get_collab_juptyer_server()
    jupyter_data = requests.get(f"http://{jupyter_address}/api/sessions").json()[0]
    filename = unquote(jupyter_data["name"])
    file_id = jupyter_data["path"].split("=")[1]
    return filename, file_id


def get_collab_path(file_id: int, drive: GoogleDrive | None = None) -> Path:
    """Use file_id to get the file path in Google Drive.

    NB needs to be mounted. User setup is conditional:

    from google.colab import drive
    drive.mount('/content/drive')
    """
    if drive is None:
        drive = auth_drive()
    if drive:
        file = drive.CreateFile({"id": file_id})
        name = file["title"]
        if file["parents"]:
            parent_id = file["parents"][0]["id"]
            return get_collab_path(parent_id, drive) / name
        else:
            return Path(name)
    raise ValueError("Google drive failed to authenticate.")


def _list_running_servers() -> Generator[dict, None, None]:
    """Iterate over the server info files of running notebook servers."""
    from jupyter_core.paths import jupyter_runtime_dir

    jupyter_runtime_dir = jupyter_runtime_dir()
    runtime_dir = Path(jupyter_runtime_dir)
    file_name = Path(f"{runtime_dir}\\{_most_recent_file(runtime_dir)[0]}")
    yield json.loads(file_name.read_bytes())


def _most_recent_file(directory_path: str | Path):
    """Get the jupyter server json file faster.

    Get the last modified file to get the currently active server.
    """
    server_most_recent_file = None
    server_most_recent_time = 0

    # kernel_most_recent_file = None
    kernel_most_recent_time = 0

    # iterate over the files in the directory using os.scandir
    for entry in os.scandir(directory_path):
        # Get the lastest modified server json
        if entry.is_file() and ("server" in entry.name and "json" in entry.name):
            # get the modification time of the file using entry.stat().st_mtime_ns
            server_mod_time = entry.stat().st_mtime_ns
            if server_mod_time > server_most_recent_time:
                # update the most recent file and its modification time
                server_most_recent_file = entry.name
                server_most_recent_time = server_mod_time
        # Get the lastest modified kernel json
        if entry.is_file() and "kernel" in entry.name:
            # get the modification time of the file using entry.stat().st_mtime_ns
            kernel_mod_time = entry.stat().st_mtime_ns
            if kernel_mod_time > kernel_most_recent_time:
                # update the most recent file and its modification time
                # kernel_most_recent_file = entry.name
                kernel_most_recent_time = kernel_mod_time

    return [server_most_recent_file]


def _get_kernel_id() -> str:
    """Returns the kernel ID of the ipykernel."""
    import ipykernel

    connection_file = Path(ipykernel.get_connection_file()).stem
    kernel_id = connection_file.split("-", 1)[1]
    return kernel_id


def _get_sessions(server: dict) -> dict:
    """Given a server, returns sessions, or HTTPError if access is denied.

    NOTE: Works only when either there is no security or there is token
    based security. An HTTPError is raised if unable to connect to a
    server.
    """
    query_str = ""
    token = server["token"]
    if token:
        query_str = f"?token={token}"
    host = "http://127.0.0.1:8888"
    url = f"/{host}api/sessions{query_str}"  # nosec
    with urllib.request.urlopen(url) as request:  # nosec
        if not (200 <= request.status_code < 300):
            raise HttpError(request.status_code, request.text, url, "GET", None)
        return json.load(request)  # nosec


def _find_notebook_path() -> tuple[dict | None, PurePath] | tuple[None, None]:  # pyright: ignore[reportUnusedFunction]
    from traitlets.config import MultipleInstanceError

    root = None
    file_path = None

    try:
        kernel_id = _get_kernel_id()
    except (MultipleInstanceError, RuntimeError):
        return None, None  # Could not determine
    server = None
    for server in _list_running_servers():
        root = server["root_dir"]
        try:
            sessions = _get_sessions(server)
            for sess in sessions:
                file_path = sess["notebook"]["path"]
                if sess["kernel"]["id"] == kernel_id:
                    return server, PurePath(sess["notebook"]["path"])  # nosec
        # There may be stale entries in the runtime directory
        except Exception:  # nosec
            pass  # nosec
    return server, PurePath(f"{root}\\{file_path}")


def auth_drive() -> GoogleDrive | None:
    """Authenticate Google Drive."""
    from google.colab import auth  # pyright: ignore[reportMissingImports]
    from google.oauth2.credentials import Credentials
    from googleapiclient.discovery import build
    from oauth2client.client import GoogleCredentials
    from pydrive2.auth import GoogleAuth
    from pydrive2.drive import GoogleDrive

    scopes = ["https://www.googleapis.com/auth/drive.metadata.readonly"]
    drive = None
    if os.getenv("GOOGLE_DRIVE_TOKEN"):
        creds = Credentials.from_authorized_user_file(os.getenv("GOOGLE_DRIVE_TOKEN"), scopes)
        drive = build("drive", "v3", credentials=creds, cache_discovery=False)
    elif os.getenv("GOOGLE_DRIVE_TOKEN") is None:
        auth.authenticate_user()
        gauth = GoogleAuth()
        gauth.credentials = GoogleCredentials.get_application_default()
        drive = GoogleDrive(gauth)
    return drive


def _get_collab_juptyer_server() -> str | None:
    process = subprocess.Popen(
        "sudo lsof -i -P -n | grep LISTEN", shell=True, stdout=subprocess.PIPE, stderr=subprocess.STDOUT
    )
    jupyter_address = None
    if process.stdout is None:
        return jupyter_address
    for line in process.stdout.readlines():
        collab_ip = re.findall(r"jupyter-n.*?TCP(.*?)\(LISTEN\)", line.decode())
        if jupyter_address:
            break
        jupyter_address = collab_ip[0].strip() if len(collab_ip) >= 1 else None
    return jupyter_address


def _find_git_path(drive: GoogleDrive | None = None):  # pyright: ignore[reportUnusedFunction]
    """Find the git path relative to the google drive directory."""
    if drive is None:
        drive = auth_drive()
    if drive:
        file_list = drive.ListFile({"q": " title='.git' "}).GetList()
        return get_collab_path(file_list[0]["id"])
    raise ValueError("Google drive failed to authenticate.")
